@@ -16,30 +16,33 @@ package App::WRT;
 # without overthinking a bunch of hair-splitting decisions and categories,
 # but whatever.  I'll try to follow it, roughly.
 
-use version; our $VERSION = version->declare("v6.3.1");
+use version; our $VERSION = version->declare("v7.0.0");
 
 use strict;
 use warnings;
 no  warnings 'uninitialized';
-use 5.10.0;
+use 5.14.0;
 use utf8;
 
+use open qw(:std :utf8);
+
 use Carp;
 use Cwd qw(getcwd abs_path);
 use Encode qw(decode encode);
 use File::Spec;
 use HTML::Entities;
 use JSON;
+use JSON::Feed;
+use Mojo::DOM;
 use XML::Atom::SimpleFeed;
 
-use App::WRT::Date;
+use App::WRT::Date qw(iso_date rfc_3339_date get_mtime month_name);
 use App::WRT::EntryStore;
 use App::WRT::FileIO;
-
-use App::WRT::HTML   qw(:all);
-use App::WRT::Image  qw(image_size);
+use App::WRT::HTML qw(:all);
+use App::WRT::Image qw(image_size);
 use App::WRT::Markup qw(line_parse image_markup eval_perl);
-use App::WRT::Util   qw(dir_list file_get_contents);
+use App::WRT::Util qw(dir_list file_get_contents);
 
 =pod
 
@@ -73,17 +76,16 @@ Using App::WRT in library form:
 
 =head1 INSTALLING
 
-It's possible but not likely this would run on a Perl as old as 5.10.0.  In
-practice, I know that it works under 5.26.2.  It should be fine on any
-reasonably modern Linux distribution, and may work on MacOS or a BSD of your
-choosing.  It's possible that it would run under the Windows Subsystem for
-Linux, but it would definitely fail under vanilla Windows; it currently makes
-too many assumptions about things like directory path separators and filesystem
-semantics.
+It's possible this would run on a Perl as old as 5.14.0.  In practice, I know
+that it works under 5.26.2.  It should be fine on any reasonably modern Linux
+distribution, and might work on BSD of your choosing.  Maybe even MacOS.  It's
+possible that it would run under the Windows Subsystem for Linux, but it would
+definitely fail under vanilla Windows; it currently makes too many assumptions
+about things like directory path separators and filesystem semantics.
 
 (Although I would like the code to be more robust across platforms, this is not
 a problem I feel much urgency about solving at the moment, since I'm pretty
-sure I am the only user of this software.  Patches would certainly be welcome.)
+sure I am the only user of this software.  Please let me know if I'm mistaken.)
 
 To install the latest development version from the main repo:
 
@@ -137,9 +139,6 @@ Contents will be generated for each year and for the entire collection of dated
 entries.  Month indices will consist of all entries for that month.  A
 top-level index file will consist of the most recent month's entries.
 
-It's possible (although not as flexible as it ought to be) to redefine the
-directory layout.  (See C<%default{entry_map}> below.)
-
 An entry may be either a plain UTF-8 text file, or a directory containing
 several such files.  If it's a directory, a file named "index" will be treated
 as the text of the entry, and all other lower-case filenames without extensions
@@ -161,13 +160,10 @@ will be treated as an icon for the corresponding entry file.
 =head2 MARKUP
 
 Entries may consist of hand-written HTML (to be passed along without further
-interpretation), a supported form of lightweight markup, or some combination
-thereof. Actually, an entry may consist of any darn thing you please, as long
-as Perl will agree that it is text, but presumably you're going to be feeding
-this to a browser.
+mangling), a supported form of lightweight markup, or some combination thereof.
 
-Header tags (<h1>, <h2>, etc.) will be used to display titles in feeds and
-other places.
+Header tags (<h1>, <h2>, etc.) will be used to display titles in feeds,
+navigation, and other places.
 
 Other special markup is indicated by a variety of HTML-like container tags.
 
@@ -200,7 +196,7 @@ root of the current wrt project:
     <include>path/to/file</include>
 
 This is a bit constraining, since it doesn't currently allow for files outside
-of the current project, but is useful for including HTML generated by an
+of the current project, but is useful for including HTML generated by some
 external script in a page.
 
 B<Several forms of lightweight markup>:
@@ -227,7 +223,9 @@ B<And a couple of shortcuts>:
      another list item
      </list>
 
-As it stands, freeverse, image, and list are not particularly robust.
+As it stands, freeverse, image, and list are not particularly robust.  In
+practice, image and list have not proven all that useful, and may be deprecated
+in a future release.
 
 =head2 TEMPLATES
 
@@ -255,9 +253,6 @@ Within templates, C<${foo}> will be replaced with the corresponding
 configuration value.  C<${content}> will always be set to the content of the
 current entry.
 
-=cut
-
-
 =head2 CONFIGURATION
 
 Configuration is read from a F<wrt.json> in the directory where the C<wrt>
@@ -295,7 +290,6 @@ Here's a verbatim copy of C<%default>, with some commentary about values.
       author         => undef,       # author name (used in template, feed)
       description    => undef,       # site description (used in template)
       content        => undef,       # place to stash content for templates
-      embedded_perl  => 1,           # evaluate embedded <perl> tags?
       default_entry  => 'new',       # what to display if no entry specified
       cache_includes => 0,           # should included files be cached in memory?
 
@@ -307,13 +301,6 @@ Here's a verbatim copy of C<%default>, with some commentary about values.
       # content in-place, like Net Neutrality protest blackouts):
       overlay        => undef,
 
-      # What gets considered an entry _path_:
-      entrypath_expr => qr/^ ([a-z0-9_\/-]+) $/x,
-
-      # What gets considered a subentry file (slightly misleading
-      # terminology here):
-      subentry_expr => qr/^[0-9a-z_-]+(\.(tgz|zip|tar[.]gz|gz|txt))?$/,
-
       # We'll show links for these, but not display them inline:
       binfile_expr   => qr/[.](tgz|zip|tar[.]gz|gz|txt|pdf)$/,
     );
@@ -334,11 +321,10 @@ my %default = (
   stylesheet_url => undef,       # path to a CSS file (used in template)
   favicon_url    => undef,       # path to a favicon (used in template)
   feed_alias     => 'feed',      # what entry path should correspond to feed?
-  feed_length    => 30,          # how many entries should there be in the feed?
+  feed_length    => 30,          # how many entries should there be in feed?
   author         => undef,       # author name (used in template, feed)
   description    => undef,       # site description (used in template)
   content        => undef,       # place to stash content for templates
-  embedded_perl  => 1,           # evaluate embedded <perl> tags?
   default_entry  => 'new',       # what to display if no entry specified
   cache_includes => 0,           # should included files be cached in memory?
 
@@ -350,61 +336,10 @@ my %default = (
   # content in-place, like Net Neutrality protest blackouts):
   overlay        => undef,
 
-  # What gets considered an entry _path_:
-  entrypath_expr => qr/^ ([a-z0-9_\/-]+) $/x,
-
-  # What gets considered a subentry file (slightly misleading
-  # terminology here):
-  subentry_expr => qr/^[0-9a-z_-]+(\.(tgz|zip|tar[.]gz|gz|txt))?$/,
-
   # We'll show links for these, but not display them inline:
   binfile_expr   => qr/[.](tgz|zip|tar[.]gz|gz|txt|pdf)$/,
 );
 
-=item $default{entry_map}
-
-A hashref which will dispatch entries matching various regexen to the
-appropriate output methods. The default looks something like this:
-
-    nnnn/[nn/nn/]doc_name - a document within a day.
-    nnnn/nn/nn            - a specific day.
-    nnnn/nn               - a month.
-    nnnn                  - a year.
-    doc_name              - a document in the root directory.
-
-You can re-map things to an arbitrary archive layout.
-
-Since the entry map is a hash, and handle() simply loops over its keys, there
-is no guaranteed precedence of patterns. Be extremely careful that no entry
-will match more than one pattern, or you will wind up with unexpected behavior.
-A good way to ensure that this does not happen is to use patterns like:
-
-    qr(
-        ^           # start of string
-        [0-9/]{4}/  # year
-        [0-9]{1,2}/ # month
-        [0-9]{1,2]  # day
-        $           # end of string
-      )x
-
-...always marking the start and end of the string explicitly.
-
-This may eventually be rewritten to use an array so that the order can be
-explicitly specified.
-
-=cut
-
-$default{entry_map} = {
-  qr'^[0-9/]{5,11}[a-z_/]+$' => sub { entry_stamped (@_, 'index') },
-
-  qr'^[0-9]{4}/[0-9]{1,2}/
-               [0-9]{1,2}$'x => sub { entry_stamped (@_, 'all'  ) },
-
-  qr'^[0-9]{4}/[0-9]{1,2}$'  => sub { month         (@_         ) },
-  qr'^[0-9]{4}$'             => sub { year          (@_         ) },
-  qr'^[a-z_]'                => sub { entry_stamped (@_, 'all'  ) },
-};
-
 =item $default{entry_descriptions}
 
 A hashref which contains a map of entry titles to entry descriptions.
@@ -450,6 +385,23 @@ sub new_from_file {
   # Grab configuration from wrt.json or other file:
   my $config_hashref = $JSON->decode(file_get_contents($config_file));
 
+  # Check for deprecated or removed configuration, and warn accordingly.
+  # TODO: These are really user-facing errors, so Carp is probably the wrong
+  #       tool for the job here.
+  if ( defined $config_hashref->{entry_map} ) {
+    carp(
+      "Caution: wrt v7.0.0 and later no longer support entry_map.\n"
+      . "Please check $config_file and remove this value."
+    );
+  }
+  if ( defined $config_hashref->{embedded_perl} ) {
+    carp(
+      "Caution: wrt v7.0.0 and later no longer support toggling embedded_perl.\n"
+      . "Please check $config_file and remove this value.\n"
+      . "Note that embedded Perl may be deprecated in a future release."
+    );
+  }
+
   # decode() returns a hashref; this needs to be dereferenced:
   return App::WRT->new(%{ $config_hashref });
 }
@@ -465,10 +417,11 @@ sub new {
   my %params = @_;
 
   # Stash absolute path to root directory.
-  # TODO: This is bad.  It's here because imgsize() winds up calling getcwd()
-  # a ton of times if you don't give it absolute paths, which is actually super
-  # inefficient.  See icon_markup() and image_markup() for usage.  image_markup()
-  # in particular is awful and should be rewritten anyway.
+  #
+  # TODO: This is bad.  It's here because imgsize() winds up calling getcwd() a
+  # ton of times if you don't give it absolute paths, which is actually super
+  # inefficient.  See icon_markup() and image_markup() for usage.
+  # image_markup() in particular is awful and should be rewritten anyway.
   $params{root_dir_abs} = abs_path($params{root_dir});
 
   my %copy_of_default = %default;
@@ -480,8 +433,6 @@ sub new {
     $self->{$p} = $params{$p};
   }
 
-  $self->{entries} = App::WRT::EntryStore->new( $self->{entry_dir} );
-
   # Check and set up template path for later use:
   $self->{template_path} = File::Spec->catfile(
     $self->{template_dir},
@@ -490,17 +441,112 @@ sub new {
   unless (-f $self->{template_path}) {
     croak($self->{template_path} . ' does not exist or is not a plain file');
   }
+  $self->{template_source} = file_get_contents($self->{template_path});
+
+  $self->{entries} = App::WRT::EntryStore->new( $self->{entry_dir} );
+
+  $self->populate_entry_caches();
+  $self->populate_metadata_cache();
 
   return $self;
 }
 
+=item populate_entry_caches()
+
+Render each renderable path, cache the HTML, and parse to extract titles.
+
+=cut
+
+sub populate_entry_caches {
+  my $self = shift;
+
+  my %html_cache;
+  my %title_cache;
+
+  foreach my $entry ($self->{entries}->all_renderable()) {
+    $html_cache{$entry} = $self->handle($entry);
+    next unless length $html_cache{$entry};
+
+    my @headers;
+    eval {
+      local $SIG{__WARN__} = sub { die; };
+      my $dom = Mojo::DOM->new($html_cache{$entry});
+      @headers = $dom->find('h1, h2, h3, h4, h5, h6')->map('text')->each;
+    };
+    if ($@) {
+      carp("Parsing issues for $entry: $@");
+    }
+    if (@headers) {
+      $title_cache{$entry} = join ' - ', @headers;
+    }
+  }
+
+  $self->{html_cache} = \%html_cache;
+  $self->{title_cache} = \%title_cache;
+}
+
+=item populate_metadata_cache()
+
+If there's any metadata, such as tagged relationships, for a given entry,
+populate an HTML blob for that stuff.
+
+XXX: Here is where we put the list of pages for a given tag, but also maybe
+other things about a page or its properties.  There should be a template /
+partial involved.
+
+=cut
+
+sub populate_metadata_cache {
+  my $self = shift;
+
+  my %metadata_html_cache;
+  foreach my $entry ($self->{entries}->all()) {
+    my $result = '';
+
+    my $tag_for_this_entry = 'tag.' . join('.', split('/', $entry));
+    my (@tagged_entries) = $self->{entries}->by_prop($tag_for_this_entry);
+    my (@alpha_entries, @dated_entries);
+    for (@tagged_entries) {
+      if (m{^\d}) {
+        push @dated_entries, $_;
+      } else {
+        push @alpha_entries, $_;
+      }
+    }
+
+    if (@tagged_entries) {
+      $result .= "<h1>entries tagged " . encode_entities($entry)
+               . "</h1>\n\n<table class=tags>";
+
+      # Things starting with letters first, then things starting with digits:
+      foreach my $tagged_entry (@alpha_entries, reverse @dated_entries) {
+        $result .= table_row(
+          table_cell(
+            a($tagged_entry, { href => $self->{url_root} . "$tagged_entry" })
+          ),
+          table_cell(
+            encode_entities($self->get_title($tagged_entry))
+          )
+        );
+        $result .= "\n";
+      }
+      $result .= "</table>";
+    }
+
+    $metadata_html_cache{$entry} = $result;
+  }
+
+  $self->{metadata_html_cache} = \%metadata_html_cache;
+}
+
 =item display($entry1, $entry2, ...)
 
 Return a string containing the given entries, which are in the form of
 date/entry strings. If no parameters are given, default to default_entry().
 
 display() expands aliases ("new" and "all", for example) as necessary, collects
-output from handle($entry), and wraps the whole thing in a template file.
+entry content and metadata from the pre-rendered HTML caches, and wraps
+everything up in the template.
 
 If C<overlay> is set, will return the value of overlay regardless of options.
 (This is useful for hackily replacing every page in a site with a single blob
@@ -518,43 +564,57 @@ sub display {
   # If no entries are defined, fall back to the default:
   $entries[0] //= $self->{default_entry};
 
-  # Title for template:
+  # Title and navigation for template:
+  $self->{page_navigation} = '';
   $self->{title} = join ' ', map { encode_entities($_) } @entries;
 
+  if (scalar @entries == 1) {
+    # We've got a single path - it could be an alias that'll expand, or it
+    # could be an individual entry.  See what can be done with navigation
+    # and title:
+    $self->{page_navigation} = $self->page_navigation($entries[0]);
+    $self->{title} = encode_entities($self->get_title($entries[0]));
+  }
+
   # Expand on any aliases:
   @entries = map { $self->expand_alias($_) } @entries;
 
-  # Hacky special case for printing the feed:
-  if ($entries[0] eq $self->{feed_alias}) {
-    return $self->feed_print(
-      $self->{entries}->recent_days( $self->{feed_length} )
-    );
-  }
-
   # To be accessed as ${content} in the template below:
-  $self->{content} = join '', map { $self->handle($_) } @entries;
-  return $self->fragment_slurp($self->{template_path});
+  $self->{content} = join '', map {
+    $self->{html_cache}{$_}
+    . '<div class=entry-metadata>'
+    . $self->{metadata_html_cache}{$_}
+    . '</div>'
+  } @entries;
+
+  return $self->line_parse(
+    $self->eval_perl($self->{template_source}),
+    $self->{template_path}
+  );
 }
 
 =item handle($entry)
 
-Return the text of an individual entry.
+Return the text of an individual entry:
+
+  nnnn/[nn/nn/]doc_name - a document within a day.
+  nnnn/nn/nn            - a specific day.
+  nnnn/nn               - a month.
+  nnnn                  - a year.
+  doc_name              - a document in the root directory.
 
 =cut
 
 sub handle {
   my ($self, $entry) = @_;
 
-  # Hashref:
-  my $map = $self->{entry_map};
-
-  # Find the first pattern in entry_map that matches this entry...
-  my ($pattern) = grep { $entry =~ $_ } keys %{ $map };
-
-  return unless defined $pattern;
-
-  # ...and use the corresponding coderef to handle the entry:
-  return $map->{$pattern}->($self, $entry);
+  for ($entry) {
+    if (m'^[\d/]+[[:lower:]_/]+$') { return entry_stamped(@_, 'index'); }
+    if (m'^\d+/\d{1,2}/\d{1,2}$' ) { return entry_stamped(@_, 'all');   }
+    if (m'^\d+/\d{1,2}$'         ) { return month(@_);                  }
+    if (m'^\d+$'                 ) { return year(@_);                   }
+    if (m'^[[:lower:]_]'         ) { return entry_stamped(@_, 'index'); }
+   }
 }
 
 =item expand_alias($option)
@@ -619,6 +679,45 @@ sub link_bar {
   return $output;
 }
 
+=item page_navigation($entry)
+
+Returns context-sensitive page navigation (next / previous links).
+
+=cut
+
+sub page_navigation {
+  my ($self) = shift;
+  my ($entry) = @_;
+  # Handle prev/next links.
+
+  if ($entry eq 'new') {
+    return qq{<a href="/all" title="all">&larr; all archives</a>};
+  }
+
+  my $output = '';
+
+  my $prev = $self->{entries}->previous($entry);
+  my $next = $self->{entries}->next($entry);
+
+  if ($prev) {
+    $output .= '<p>previous: <a title="previous" href="'
+             . encode_entities($self->{url_root} . $prev)
+             . '">'
+             . encode_entities($self->get_title($prev))
+             . '</a></p> ';
+  }
+
+  if ($next) {
+    $output .= '<p>next: <a title="next" href="'
+             . encode_entities($self->{url_root} . $next)
+             . '">'
+             . encode_entities($self->get_title($next))
+             . '</a></p>';
+  }
+
+  return $output;
+}
+
 =item year($year)
 
 List out the updates for a year.
@@ -671,7 +770,7 @@ sub year {
     my ($month_file, $month_url) = $self->root_locations($month);
     my $link = a(
       { href => "${month_url}/" },
-      App::WRT::Date::month_name($self->{entries}->basename($month))
+      month_name($self->{entries}->basename($month))
     );
 
     $year_text .= table_row(
@@ -741,25 +840,29 @@ sub entry_stamped {
   );
 }
 
-=item entry_topic_list($entry)
+=item entry_tag_list($entry)
 
-Get a list of topics (by tag-* files) for the entry.  This hardcodes part of a
-p1k3-specific thing which should be moved into wrt entirely.
+Get tag links for the entry.
 
 =cut
 
-sub entry_topic_list {
+sub entry_tag_list {
   my $self = shift;
   my ($entry) = @_;
 
-  my @tags = sort grep { m/^tag-.*/ } $self->{entries}->props_for($entry);
+  my @tags = sort grep {
+    m/^tag [.] .*/x
+  } $self->{entries}->props_for($entry);
 
-  return '' unless @tags;
+  if (@tags) {
+    return 'tags: ' . join ', ', map {
+      s/^tag[.](.*)$/$1/;
+      s{[.]}{/}g;
+      a(encode_entities($_), { href => $self->{url_root} . $_ })
+    } @tags;
+  }
 
-  return join ', ', map {
-    s/^tag-(.*)$/$1/;
-    a($_, { href => $self->{url_root} . 'topics/' . $_ })
-  } @tags;
+  return '';
 }
 
 =item entry($entry)
@@ -795,7 +898,7 @@ sub entry {
   }
 
   # Followed by any sub-entries:
-  my @sub_entries = $self->get_sub_entries($entry_loc);
+  my @sub_entries = $self->{entries}->get_sub_entries($entry);
 
   if (@sub_entries >= 1) {
     # If the wrt-noexpand property is present, then don't expand sub-entries.
@@ -824,23 +927,6 @@ sub entry {
   return $result;
 }
 
-=item get_sub_entries($entry_loc)
-
-Returns "sub entries" based on the C<subentry_expr> regexp.
-
-=cut
-
-sub get_sub_entries {
-  my ($self, $entry_loc) = @_;
-
-  # index gets special treatment as the text body of an entry, rather
-  # than as a sub-entry:
-  my %ignore = ('index' => 1);
-
-  return grep { ! $ignore{$_} }
-              dir_list($entry_loc, 'alpha', $self->{subentry_expr});
-}
-
 =item list_contents($entry, @entries)
 
 Returns links (maybe with icons) for a set of sub-entries within an entry.
@@ -866,6 +952,26 @@ sub list_contents {
   return p( em('more:') . " $contents" ) . "\n";
 }
 
+=item get_title($entry)
+
+Returns a title for the entry - potentially a cached one extracted earlier from
+the entry's HTML; otherwise just reuse the entry path itself.
+
+=cut
+
+sub get_title {
+  my ($self, $entry) = @_;
+
+  # Base title - just the entry path:
+  my $title = $entry;
+
+  # Do we have anything in the cache?
+  if (defined $self->{title_cache}{$entry}) {
+    $title = $self->{title_cache}{$entry};
+  }
+  return $title;
+}
+
 =item icon_markup($entry, $alt)
 
 Check if an icon exists for a given entry if so, return markup to include it.
@@ -891,6 +997,9 @@ sub icon_markup {
   }
   elsif ($self->{entries}->is_dir($entry)) {
     $icon_basepath = "$entry/index.icon";
+  } else {
+    # XXX there are bugs lurking here for virtual entries probably
+    return;
   }
 
   # First suffix found will be used:
@@ -943,11 +1052,13 @@ sub datestamp {
              title => $fragment }, $fragment);
   }
 
-  my $stamp = $self->entry_topic_list($entry)
-            . " :: "
-            . join(" /\n", @fragment_stamps);
+  my $stamp = join(" /\n", @fragment_stamps);
+  my $tag_list = $self->entry_tag_list($entry);
+  if ($tag_list) {
+    $stamp .= "<br>\n$tag_list";
+  }
 
-  return p({class => 'datelink'}, "\n$stamp\n");
+  return p({class => 'datestamp'}, "\n<i>$stamp</i>\n");
 }
 
 =item fragment_slurp($file)
@@ -960,14 +1071,11 @@ a string. Takes one parameter, the name of the file.
 
 sub fragment_slurp {
   my $self = shift;
-
   my ($file) = @_;
-  my $everything = file_get_contents($file);
 
   return $self->line_parse(
-    # Handle embedded perl first
-    ($self->{embedded_perl} ? $self->eval_perl($everything) : $everything),
-    $file # some context to work with
+    $self->eval_perl(file_get_contents($file)),
+    $file
   );
 }
 
@@ -985,6 +1093,39 @@ sub root_locations {
   );
 }
 
+=item feed_print_recent($count)
+
+Print $count recent entries, falling back to the configured $feed_length.
+
+=cut
+
+sub feed_print_recent {
+  my ($self, $count) = @_;
+
+  $count //= $self->{feed_length};
+
+  return $self->feed_print(
+    $self->{entries}->recent_days($count)
+  );
+}
+
+=item feed_print_json_recent($count)
+
+Print $count recent entries in JSON, falling back to the configured
+$feed_length.
+
+=cut
+
+sub feed_print_json_recent {
+  my ($self, $count) = @_;
+
+  $count //= $self->{feed_length};
+
+  return $self->feed_print_json(
+    $self->{entries}->recent_days($count)
+  );
+}
+
 =item feed_print(@entries)
 
 Return an Atom feed for the given list of entries.
@@ -1023,12 +1164,9 @@ sub feed_print {
 
   my ($first_entry_file, $first_entry_url) = $self->root_locations($entries[0]);
 
-  # TODO: Probably ought to consider utf-8 in titles, authors, etc., as well as
-  # entry content.
-
   my $feed = XML::Atom::SimpleFeed->new(
     -encoding => 'UTF-8',
-    title     => $self->{title_prefix} . '::' . $self->{title},
+    title     => $self->{title_prefix} . "::" . $self->{feed_alias},
     subtitle  => $self->{description},
     link      => $self->{url_root},
     link      => { rel => 'self', href => $feed_url, },
@@ -1036,38 +1174,88 @@ sub feed_print {
     author    => $self->{author},
     id        => $self->{url_root},
     generator => 'App::WRT.pm / XML::Atom::SimpleFeed',
-    updated   => App::WRT::Date::iso_date(App::WRT::Date::get_mtime($first_entry_file)),
+    updated   => iso_date(get_mtime($first_entry_file)),
   );
 
   foreach my $entry (@entries) {
-    my $content = $self->entry($entry) . "\n" . $self->datestamp($entry);
-    my $utf8_content = decode('UTF-8', $content, Encode::FB_CROAK);
-
-    my $title = $entry;
-    my ($entry_file, $entry_url) = $self->root_locations($entry);
-
-    # Try to pull out a header:
-    my ($extracted_title) = $utf8_content =~ m{<h1.*?>(.*?)</h1>}s;
-    my (@subtitles)       = $utf8_content =~ m{<h2.*?>(.*?)</h2>}sg;
-
-    if ($extracted_title) {
-      $title = $extracted_title;
-      if (@subtitles) {
-        $title .= ' - ' . join ' - ', @subtitles;
-      }
+    my $content = $self->{html_cache}{$entry};
+    if ( $self->{metadata_html_cache}{$entry} ) {
+      $content .= '<div class=entry-metadata>'
+                . $self->{metadata_html_cache}{$entry}
+                . '</div>';
     }
 
+    my ($entry_file, $entry_url) = $self->root_locations($entry);
+
     $feed->add_entry(
-      title   => $title,
+      title   => $self->get_title($entry),
       link    => $entry_url,
       id      => $entry_url,
-      content => $utf8_content,
-      updated => App::WRT::Date::iso_date(App::WRT::Date::get_mtime($entry_file)),
+      content => $content,
+      updated => iso_date(get_mtime($entry_file)),
     );
   }
 
-  # return "Content-type: application/atom+xml\n\n" . $feed->as_string;
-  return $feed->as_string;
+  # Note: This output should be served with
+  # Content-type: application/atom+xml
+  #
+  # I'm not, to be frank, entirely clear on why the decode() call here is
+  # necessary:
+  return decode('UTF-8', $feed->as_string);
+}
+
+=item feed_print_json
+
+Like feed_print(), but for JSON Feed.
+
+L<https://jsonfeed.org/>
+
+=cut
+
+sub feed_print_json {
+  my ($self, @entries) = @_;
+  my ($first_entry_file, $first_entry_url) = $self->root_locations($entries[0]);
+
+  my $json_feed_url = $self->{url_root} . $self->{feed_alias} . '.json';
+
+  my $user_comment = "This feed allows you to read the posts from this site in"
+    . " any feed reader that supports the JSON Feed format. To "
+    . "add this feed to your reader, copy the following URL — "
+    . "$json_feed_url — and add it your reader.";
+
+  my $feed = JSON::Feed->new(
+    user_comment  => $user_comment,
+    title         => $self->{title_prefix} . "::" . $self->{feed_alias},
+    home_page_url => $self->{url_root},
+    feed_url      => $json_feed_url,
+    description   => $self->{description},
+    author        => +{ name => $self->{author}, },
+  );
+
+  if (defined $self->{favicon_url}) {
+    $feed->set('favicon', $self->{favicon_url});
+  }
+
+  foreach my $entry (@entries) {
+    my $content = $self->{html_cache}{$entry};
+    if ($self->{metadata_html_cache}{$entry}) {
+      $content .= '<div class=entry-metadata>'
+                . $self->{metadata_html_cache}{$entry}
+                . '</div>';
+    }
+
+    my ($entry_file, $entry_url) = $self->root_locations($entry);
+
+    $feed->add_item(
+      id            => $entry_url,
+      title         => $self->get_title($entry),
+      content_html  => $content,
+      date_modified => rfc_3339_date(get_mtime($entry_file)),
+    );
+  }
+
+  # Output
+  return $feed->to_string;
 }
 
 =back
